# fourFn.py
#
# Demonstration of the parsing module, implementing a simple 4-function expression parser,
# with support for scientific notation, and symbols for e and pi
#
# Copyright 2003, by Paul McGuire
#
from pyparsing import Literal,CaselessLiteral,Word,Combine,Group,Optional,ZeroOrMore,Forward,nums

exprStack = []

def pushFirst( str, loc, toks ):
    global exprStack
    if toks: 
        exprStack.append( toks[0] )
    return toks

bnf = None
def BNF():
    global bnf
    if not bnf:
        point = Literal( "." )
        e     = CaselessLiteral( "E" )
        fnumber = Combine( Word( "+-"+nums, nums ) + 
                           Optional( point + Optional( Word( nums ) ) ) +
                           Optional( e + Word( "+-"+nums, nums ) ) )
     
        plus  = Literal( "+" )
        minus = Literal( "-" )
        mult  = Literal( "*" )
        div   = Literal( "/" )
        lpar  = Literal( "(" ).suppress()
        rpar  = Literal( ")" ).suppress()
        addop  = plus | minus
        multop = mult | div
        pi    = CaselessLiteral( "PI" )
        
        expr = Forward()
        factor = ( pi | e | fnumber ).setParseAction( pushFirst ) | ( lpar + expr.suppress() + rpar )
        term = factor + ZeroOrMore( ( multop + factor ).setParseAction( pushFirst ) )
        expr << term + ZeroOrMore( ( addop + term ).setParseAction( pushFirst ) )
        bnf = expr
    return bnf

# map operator symbols to corresponding arithmetic operations
opn = { "+" : ( lambda a,b: a + b ),
        "-" : ( lambda a,b: a - b ),
        "*" : ( lambda a,b: a * b ),
        "/" : ( lambda a,b: a / b ) }
def evaluateStack( s ):
    op = s.pop()
    if op in "+-*/":
        op2 = evaluateStack( s )
        op1 = evaluateStack( s )
        return opn[op]( op1, op2 )
    elif op == "PI":
        return 3.1415926535
    elif op == "E":
        return 2.718281828
    else:
        return float( op )

if __name__ == "__main__":
  
    def test( str ):
        global exprStack
        exprStack = []
        results = BNF().parseString( str )
        print str, "->", results, "=>", exprStack, "=", evaluateStack( exprStack )
  
    test( "9" )
    test( "9 + 3 + 6" )
    test( "9 + 3 / 11" )
    test( "(9 + 3)" )
    test( "(9+3) / 11" )
    test( "9 - (12 - 6)" )
    test( "2*3.14159" )
    test( "3.1415926535*3.1415926535 / 10" )
    test( "PI * PI / 10" )
    test( "PI*PI/10" )
    test( "6.02E23 * 8.048" )
    test( "e / 3" )


"""
Test output:
>pythonw -u fourFn.py
9 -> ['9'] => ['9'] = 9.0
9 + 3 + 6 -> ['9', '+', '3', '+', '6'] => ['9', '3', '+', '6', '+'] = 18.0
9 + 3 / 11 -> ['9', '+', '3', '/', '11'] => ['9', '3', '11', '/', '+'] = 9.27272727273
(9 + 3) -> ['(', '9', '+', '3', ')'] => ['9', '3', '+'] = 12.0
(9+3) / 11 -> ['(', '9', '+', '3', ')', '/', '11'] => ['9', '3', '+', '11', '/'] = 1.09090909091
9 - (12 - 6) -> ['9', '-', '(', '12', '-', '6', ')'] => ['9', '12', '6', '-', '-'] = 3.0
2*3.14159 -> ['2', '*', '3.14159'] => ['2', '3.14159', '*'] = 6.28318
3.1415926535*3.1415926535 / 10 -> ['3.1415926535', '*', '3.1415926535', '/', '10'] => ['3.1415926535', '3.1415926535', '*', '10', '/'] = 0.986960440053
PI * PI / 10 -> ['PI', '*', 'PI', '/', '10'] => ['PI', 'PI', '*', '10', '/'] = 0.986960440053
PI*PI/10 -> ['PI', '*', 'PI', '/', '10'] => ['PI', 'PI', '*', '10', '/'] = 0.986960440053
6.02E23 * 8.048 -> ['6.02E23', '*', '8.048'] => ['6.02E23', '8.048', '*'] = 4.844896e+024
e / 3 -> ['E', '/', '3'] => ['E', '3', '/'] = 0.906093942667
>Exit code: 0
"""